from abc import abstractmethod
from typing import Any, Optional, Sequence, Tuple, Type

import turu.core.async_cursor
from turu.core.protocols.async_connection import AsyncConnectionProtocol
from turu.core.protocols.async_cursor import Parameters
from typing_extensions import Never, Self


class AsyncConnection(AsyncConnectionProtocol):
    @classmethod
    @abstractmethod
    async def connect(cls, *args: Any, **kwargs: Any) -> Self:
        """Connect to a database."""
        ...

    @classmethod
    @abstractmethod
    async def connect_from_env(cls, *args: Any, **kwargs: Any) -> Self:
        """Connect to a database using environment variables."""
        ...

    @abstractmethod
    async def close(self) -> None:
        ...

    @abstractmethod
    async def commit(self) -> None:
        ...

    @abstractmethod
    async def rollback(self) -> None:
        ...

    @abstractmethod
    async def cursor(self) -> turu.core.async_cursor.AsyncCursor[Never, Any]:
        ...

    async def execute(
        self,
        operation: str,
        parameters: Optional[Parameters] = None,
        /,
    ) -> turu.core.async_cursor.AsyncCursor[Tuple[Any], Parameters]:
        """Prepare and execute a database operation (query or command).

        This is not defined in [PEP 249](https://peps.python.org/pep-0249/),
        but is simply a convenient shortcut to `.cursor().execute()`.

        Parameters:
            operation: A database operation (query or command).
            parameters: Parameters may be provided as sequence or mapping and will be bound to variables in the operation.

        Returns:
            A cursor that holds a reference to an operation.
        """

        return await (await self.cursor()).execute(operation, parameters)

    async def executemany(
        self,
        operation: str,
        seq_of_parameters: Sequence[Parameters],
        /,
    ) -> turu.core.async_cursor.AsyncCursor[Tuple[Any], Parameters]:
        """Prepare a database operation (query or command)
        and then execute it against all parameter sequences or mappings.

        This is not defined in [PEP 249](https://peps.python.org/pep-0249/),
        but is simply a convenient shortcut to `.cursor().executemany()`.

        Parameters:
            operation: A database operation (query or command).
            seq_of_parameters: Parameters may be provided as sequence or mapping and will be bound to variables in the operation.

        Returns:
            A cursor that holds a reference to an operation.
        """

        return await (await self.cursor()).executemany(operation, seq_of_parameters)

    async def execute_map(
        self,
        row_type: Type[turu.core.async_cursor.GenericNewRowType],
        operation: str,
        parameters: Optional[Parameters] = None,
        /,
    ) -> turu.core.async_cursor.AsyncCursor[
        turu.core.async_cursor.GenericNewRowType, Parameters
    ]:
        """
        Execute a database operation (query or command) and map each row to a `row_type`.

        This is not defined in [PEP 249](https://peps.python.org/pep-0249/),
        but is simply a convenient shortcut to `.cursor().execute_map()`.

        Parameters:
            row_type: The type of the row that will be returned.
            operation: A database operation (query or command).
            parameters: Parameters may be provided as sequence or mapping and will be bound to variables in the operation.

        Returns:
            A cursor that holds a reference to an operation.
        """

        return await (await self.cursor()).execute_map(row_type, operation, parameters)

    async def executemany_map(
        self,
        row_type: Type[turu.core.async_cursor.GenericNewRowType],
        operation: str,
        seq_of_parameters: Sequence[Parameters],
        /,
    ) -> turu.core.async_cursor.AsyncCursor[
        turu.core.async_cursor.GenericNewRowType, Parameters
    ]:
        """Execute a database operation (query or command) against all parameter sequences or mappings.

        This is not defined in [PEP 249](https://peps.python.org/pep-0249/),
        but is simply a convenient shortcut to `.cursor().executemany_map()`.

        Parameters:
            row_type: The type of the row that will be returned.
            operation: A database operation (query or command).
            seq_of_parameters: Parameters may be provided as sequence or mapping and will be bound to variables in the operation.

        Returns:
            A cursor that holds a reference to an operation.
        """

        return await (await self.cursor()).executemany_map(
            row_type, operation, seq_of_parameters
        )



from typing import Any, List, Optional, Sequence, Tuple, Type, Union, overload

from turu.core.cursor import GenericRowType
from turu.core.mock.exception import (
    TuruMockResponseTypeMismatchError,
    TuruMockStoreDataNotFoundError,
)


class TuruMockStore:
    def __init__(self):
        self._data: List[Tuple[Optional[Type], Union[Sequence, None, Exception]]] = []
        self._counter = 0

    @overload
    def inject_response(
        self,
        row_type: None,
        response: Union[GenericRowType, Sequence[GenericRowType], None, Exception],
    ):
        ...

    @overload
    def inject_response(
        self,
        row_type: Type[GenericRowType],
        response: Union[GenericRowType, Sequence[GenericRowType], Exception],
    ):
        ...

    def inject_response(self, row_type, response):
        if row_type is not None and isinstance(response, row_type):
            response = (response,)

        self._data.append((row_type, response))

    def provide_response(
        self,
        row_type: Optional[Type],
    ) -> "Optional[Sequence[Any]]":
        self._counter += 1

        if len(self._data) == 0:
            raise TuruMockStoreDataNotFoundError(self._counter)

        _row_type, _response = self._data.pop(0)

        if _row_type is not row_type and not (
            row_type.__module__ == _row_type.__module__
            and row_type.__name__ == _row_type.__name__
        ):
            raise TuruMockResponseTypeMismatchError(row_type, _row_type, self._counter)

        if isinstance(_response, Exception):
            raise _response

        return _response



from typing_extensions import TypeAlias


class _NotSupportFeature:
    pass


try:
    import pydantic  # type: ignore[import]

    USE_PYDANTIC = True
    PydanticModel: TypeAlias = pydantic.BaseModel  # type: ignore

except ImportError:
    USE_PYDANTIC = False

    PydanticModel: TypeAlias = _NotSupportFeature  # type: ignore



from abc import abstractmethod
from typing import Any, Optional, Sequence, Tuple, Type

import turu.core.cursor
from turu.core.protocols.connection import ConnectionProtocol
from turu.core.protocols.cursor import Parameters
from typing_extensions import Never, Self, override


class Connection(ConnectionProtocol):
    @classmethod
    @abstractmethod
    def connect(cls, *args: Any, **kwargs: Any) -> Self:
        """Connect to a database."""
        ...

    @classmethod
    @abstractmethod
    def connect_from_env(cls, *args: Any, **kwargs: Any) -> Self:
        """Connect to a database using environment variables."""
        ...

    @abstractmethod
    @override
    def close(self) -> None:
        ...

    @abstractmethod
    @override
    def commit(self) -> None:
        ...

    @abstractmethod
    @override
    def rollback(self) -> None:
        ...

    @abstractmethod
    @override
    def cursor(self) -> turu.core.cursor.Cursor[Never, Any]:
        ...

    def execute(
        self,
        operation: str,
        parameters: Optional[Parameters] = None,
        /,
    ) -> turu.core.cursor.Cursor[Tuple[Any], Parameters]:
        """Prepare and execute a database operation (query or command).

        This is not defined in [PEP 249](https://peps.python.org/pep-0249/),
        but is simply a convenient shortcut to `.cursor().execute()`.

        Parameters:
            operation: A database operation (query or command).
            parameters: Parameters may be provided as sequence or mapping and will be bound to variables in the operation.

        Returns:
            A cursor that holds a reference to an operation.
        """

        return self.cursor().execute(operation, parameters)

    def executemany(
        self,
        operation: str,
        seq_of_parameters: Sequence[Parameters],
        /,
    ) -> turu.core.cursor.Cursor[Tuple[Any], Parameters]:
        """Prepare a database operation (query or command)
        and then execute it against all parameter sequences or mappings.

        This is not defined in [PEP 249](https://peps.python.org/pep-0249/),
        but is simply a convenient shortcut to `.cursor().executemany()`.

        Parameters:
            operation: A database operation (query or command).
            seq_of_parameters: Parameters may be provided as sequence or mapping and will be bound to variables in the operation.

        Returns:
            A cursor that holds a reference to an operation.
        """

        return self.cursor().executemany(operation, seq_of_parameters)

    def execute_map(
        self,
        row_type: Type[turu.core.cursor.GenericNewRowType],
        operation: str,
        parameters: Optional[Parameters] = None,
        /,
    ) -> turu.core.cursor.Cursor[turu.core.cursor.GenericNewRowType, Parameters]:
        """
        Execute a database operation (query or command) and map each row to a `row_type`.

        This is not defined in [PEP 249](https://peps.python.org/pep-0249/),
        but is simply a convenient shortcut to `.cursor().execute_map()`.

        Parameters:
            row_type: The type of the row that will be returned.
            operation: A database operation (query or command).
            parameters: Parameters may be provided as sequence or mapping and will be bound to variables in the operation.

        Returns:
            A cursor that holds a reference to an operation.
        """

        return self.cursor().execute_map(row_type, operation, parameters)

    def executemany_map(
        self,
        row_type: Type[turu.core.cursor.GenericNewRowType],
        operation: str,
        seq_of_parameters: Sequence[Parameters],
        /,
    ) -> turu.core.cursor.Cursor[turu.core.cursor.GenericNewRowType, Parameters]:
        """Execute a database operation (query or command) against all parameter sequences or mappings.

        This is not defined in [PEP 249](https://peps.python.org/pep-0249/),
        but is simply a convenient shortcut to `.cursor().executemany_map()`.

        Parameters:
            row_type: The type of the row that will be returned.
            operation: A database operation (query or command).
            seq_of_parameters: Parameters may be provided as sequence or mapping and will be bound to variables in the operation.

        Returns:
            A cursor that holds a reference to an operation.
        """

        return self.cursor().executemany_map(row_type, operation, seq_of_parameters)



from typing import Any, Protocol, TextIO


class RecorderProtcol(Protocol):
    @property
    def file(self) -> TextIO:
        ...

    def record(self, rows: Any) -> None:
        ...

    def close(self) -> None:
        ...



from typing import Optional

from turu.core.exception import TuruError


class TuruMockError(TuruError):
    pass


class TuruMockStoreDataNotFoundError(TuruMockError):
    def __init__(self, counter: int) -> None:
        self.counter = counter

    @property
    def message(self) -> str:
        return f"Mock data not found. counter: {self.counter}"


class TuruMockResponseTypeMismatchError(TuruMockError):
    def __init__(
        self, expected: Optional[type], actual: Optional[type], counter: int
    ) -> None:
        self.expected = expected
        self.actual = actual
        self.counter = counter

    @property
    def message(self) -> str:
        expected_type = self.expected.__name__ if self.expected else None
        actual_type = self.actual.__name__ if self.actual else None

        return f"Mock response type mismatch: Expected {expected_type}, got {actual_type}. counter: {self.counter}"


class TuruMockUnexpectedFetchError(TuruMockError):
    @property
    def message(self) -> str:
        return "Mock fetch is unexpected. use execute_map() or executemany_map() to specify row type."


class TuruCsvHeaderOptionRequiredError(TuruError):
    def __init__(self, row_type: type):
        self.row_type = row_type

    @property
    def message(self) -> str:
        return f'"{self.row_type}" requires header=True when reading from CSV'



from abc import abstractmethod
from dataclasses import is_dataclass
from typing import (
    Any,
    Generic,
    List,
    Optional,
    Sequence,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
)

from turu.core.exception import TuruRowTypeMismatchError
from turu.core.features import USE_PYDANTIC, PydanticModel
from turu.core.protocols.cursor import CursorProtocol, Parameters
from turu.core.protocols.dataclass import Dataclass
from typing_extensions import Self, override

RowType = Union[Tuple[Any], Dataclass, PydanticModel]
GenericRowType = TypeVar("GenericRowType", bound=RowType)
GenericNewRowType = TypeVar("GenericNewRowType", bound=RowType)


class Cursor(Generic[GenericRowType, Parameters], CursorProtocol[Parameters]):
    @property
    @abstractmethod
    def rowcount(self) -> int:
        ...

    @property
    @abstractmethod
    def arraysize(self) -> int:
        ...

    @arraysize.setter
    @abstractmethod
    def arraysize(self, size: int) -> None:
        ...

    @abstractmethod
    def close(self) -> None:
        ...

    @abstractmethod
    def execute(
        self, operation: str, parameters: Optional[Parameters] = None, /
    ) -> "Cursor":
        ...

    @abstractmethod
    def executemany(
        self, operation: str, seq_of_parameters: Sequence[Parameters], /
    ) -> "Cursor":
        ...

    @abstractmethod
    def execute_map(
        self,
        row_type: Type[GenericNewRowType],
        operation: str,
        parameters: Optional[Parameters] = None,
        /,
    ) -> "Cursor[GenericNewRowType, Parameters]":
        """
        Execute a database operation (query or command) and map each row to a `row_type`.

        Parameters:
            row_type: The type of the row that will be returned.
            operation: A database operation (query or command).
            parameters: Parameters may be provided as sequence or mapping and will be bound to variables in the operation.

        Returns:
            A cursor that holds a reference to an operation.
        """
        ...

    @abstractmethod
    def executemany_map(
        self,
        row_type: Type[GenericNewRowType],
        operation: str,
        seq_of_parameters: Sequence[Parameters],
        /,
    ) -> "Cursor[GenericNewRowType, Parameters]":
        """Execute a database operation (query or command) against all parameter sequences or mappings.

        Parameters:
            row_type: The type of the row that will be returned.
            operation: A database operation (query or command).
            seq_of_parameters: Parameters may be provided as sequence or mapping and will be bound to variables in the operation.

        Returns:
            A cursor that holds a reference to an operation.
        """
        ...

    @abstractmethod
    def fetchone(self) -> Optional[GenericRowType]:
        ...

    @abstractmethod
    def fetchmany(self, size: Optional[int] = None) -> List[GenericRowType]:
        ...

    @abstractmethod
    def fetchall(self) -> List[GenericRowType]:
        ...

    @override
    def __iter__(self) -> Self:
        return self

    @abstractmethod
    def __next__(self) -> GenericRowType:
        ...

    @override
    def __enter__(self):
        return self

    @override
    def __exit__(self, exc_type, exc_value, traceback):
        self.close()


GenericCursor = TypeVar("GenericCursor", bound=Cursor)


def map_row(row_type: Optional[Type[GenericRowType]], row: Any) -> GenericRowType:
    if row_type is None:
        return row

    if is_dataclass(row_type):
        return row_type(
            **{
                key: data
                for key, data in zip(row_type.__dataclass_fields__.keys(), row)
            }
        )  # type: ignore

    elif issubclass(row_type, tuple):
        return row_type._make(row)  # type: ignore

    elif USE_PYDANTIC and issubclass(row_type, PydanticModel):
        return row_type(
            **{
                key: data
                for key, data in zip(
                    cast(PydanticModel, row_type).model_fields.keys(), row
                )
            }
        )  # type: ignore

    raise TuruRowTypeMismatchError(row_type, row.__class__)



from .async_connection import MockAsyncConnection
from .async_cursor import MockAsyncCursor
from .connection import MockConnection
from .cursor import MockCursor

__all__ = [
    "MockAsyncConnection",
    "MockAsyncCursor",
    "MockConnection",
    "MockCursor",
]



import csv
import pathlib
from abc import abstractmethod
from typing import (
    Any,
    Optional,
    Sequence,
    Type,
    Union,
    overload,
)

import turu.core.async_connection
from turu.core.cursor import GenericRowType, map_row
from turu.core.mock.connection import CSVOptions
from turu.core.mock.store import TuruMockStore
from typing_extensions import Never, Self, Unpack

from .async_cursor import MockAsyncCursor


class MockAsyncConnection(turu.core.async_connection.AsyncConnection):
    def __init__(self, store: Optional[TuruMockStore] = None):
        self._turu_mock_store = store or TuruMockStore()

    @classmethod
    async def connect(cls, *args: Any, **kwargs: Any) -> Self:
        return cls()

    @classmethod
    async def connect_from_env(cls, *args: Any, **kwargs: Any) -> Self:
        return cls()

    def chain(self) -> Self:
        """this method is just for code formatting by black."""

        return self

    @overload
    def inject_response(
        self,
        row_type: None,
        response: Union[Optional[Sequence[Any]], Exception] = None,
    ) -> Self:
        ...

    @overload
    def inject_response(
        self,
        row_type: Type[GenericRowType],
        response: Union[Sequence[GenericRowType], Exception],
    ) -> Self:
        ...

    def inject_response(
        self,
        row_type,
        response=None,
    ):
        self._turu_mock_store.inject_response(row_type, response)
        return self

    @overload
    def inject_response_from_csv(
        self,
        row_type: None,
        filepath: Union[str, pathlib.Path],
        **options: Unpack[CSVOptions],
    ) -> Self:
        ...

    @overload
    def inject_response_from_csv(
        self,
        row_type: Type[GenericRowType],
        filepath: Union[str, pathlib.Path],
        **options: Unpack[CSVOptions],
    ) -> Self:
        ...

    def inject_response_from_csv(
        self,
        row_type: Optional[Type[GenericRowType]],
        filepath: Union[str, pathlib.Path],
        **options: Unpack[CSVOptions],
    ):
        with open(filepath, "r") as file:
            reader = csv.reader(file)

            if options.get("header", True):
                next(reader)

            response = [map_row(row_type, row) for row in reader]

        self.inject_response(row_type, response)

        return self

    async def close(self) -> None:
        pass

    async def commit(self) -> None:
        pass

    async def rollback(self) -> None:
        pass

    @abstractmethod
    async def cursor(self) -> MockAsyncCursor[Never, Any]:
        ...
